@@ -1,7 +1,7 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
 #
-# Copyright (c) 2012-2020 Snowflake Computing Inc. All right reserved.
+# Copyright (c) 2012-2021 Snowflake Computing Inc. All right reserved.
 #
 
 import codecs
@@ -17,12 +17,14 @@ from os.path import expanduser
 from threading import Lock, Thread
 
 from .auth_keypair import AuthByKeyPair
+from .auth_usrpwdmfa import AuthByUsrPwdMfa
 from .compat import IS_LINUX, IS_MACOS, IS_WINDOWS, urlencode
 from .constants import (
     HTTP_HEADER_ACCEPT,
     HTTP_HEADER_CONTENT_TYPE,
     HTTP_HEADER_SERVICE_NAME,
     HTTP_HEADER_USER_AGENT,
+    PARAMETER_CLIENT_REQUEST_MFA_TOKEN,
     PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL,
 )
 from .description import COMPILER, IMPLEMENTATION, OPERATING_SYSTEM, PLATFORM, PYTHON_VERSION
@@ -79,6 +81,9 @@ KEYRING_SERVICE_NAME = "net.snowflake.temporary_token"
 KEYRING_USER = "temp_token"
 KEYRING_DRIVER_NAME = "SNOWFLAKE-PYTHON-DRIVER"
 
+ID_TOKEN = "ID_TOKEN"
+MFA_TOKEN = "MFATOKEN"
+
 
 class Auth(object):
     """Snowflake Authenticator."""
@@ -91,8 +96,7 @@ class Auth(object):
                        internal_application_name,
                        internal_application_version,
                        ocsp_mode, login_timeout,
-                       network_timeout=None,
-                       store_temp_cred=None):
+                       network_timeout=None):
         return {
             'data': {
                 "CLIENT_APP_ID": internal_application_name,
@@ -111,7 +115,6 @@ class Auth(object):
                     "TRACING": logger.getEffectiveLevel(),
                     "LOGIN_TIMEOUT": login_timeout,
                     "NETWORK_TIMEOUT": network_timeout,
-                    "CLIENT_STORE_TEMPORARY_CREDENTIAL": store_temp_cred,
                 }
             },
         }
@@ -137,12 +140,6 @@ class Auth(object):
             headers[HTTP_HEADER_SERVICE_NAME] = \
                 session_parameters[HTTP_HEADER_SERVICE_NAME]
         url = "/session/v1/login-request"
-        if session_parameters is not None \
-                and PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL in session_parameters:
-            store_temp_cred = session_parameters[
-                PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL]
-        else:
-            store_temp_cred = None
 
         body_template = Auth.base_auth_data(
             user, account, self._rest._connection.application,
@@ -151,7 +148,6 @@ class Auth(object):
             self._rest._connection._ocsp_mode(),
             self._rest._connection._login_timeout,
             self._rest._connection._network_timeout,
-            store_temp_cred,
         )
 
         body = copy.deepcopy(body_template)
@@ -297,9 +293,9 @@ class Auth(object):
             errno = ret.get('code', ER_FAILED_TO_CONNECT_TO_DB)
             if errno == ID_TOKEN_INVALID_LOGIN_REQUEST_GS_CODE:
                 # clear stored id_token if failed to connect because of id_token
-                # raise an exception for reauthing without id_token
+                # raise an exception for reauth without id_token
                 self._rest.id_token = None
-                delete_temporary_credential(self._rest._host, user, True)
+                delete_temporary_credential(self._rest._host, user, ID_TOKEN)
                 raise ReauthenticationRequest(
                     ProgrammingError(
                         msg=ret['message'],
@@ -314,6 +310,8 @@ class Auth(object):
                     str(auth_instance._jwt_token_exp),
                     str(datetime.utcnow())
                 )
+            if type(auth_instance) is AuthByUsrPwdMfa:
+                delete_temporary_credential(self._rest._host, user, MFA_TOKEN)
             Error.errorhandler_wrapper(
                 self._rest._connection, None, DatabaseError,
                 {
@@ -336,19 +334,20 @@ class Auth(object):
                          'NULL')
             logger.debug('id_token = %s',
                          '******' if ret['data'].get(
-                             'id_token') is not None else
+                             'idToken') is not None else
+                         'NULL')
+            logger.debug('mfa_token = %s',
+                         '******' if ret['data'].get(
+                             'mfaToken') is not None else
                          'NULL')
             self._rest.update_tokens(
                 ret['data']['token'], ret['data']['masterToken'],
                 master_validity_in_seconds=ret['data'].get(
                     'masterValidityInSeconds'),
-                id_token=ret['data'].get('idToken')
+                id_token=ret['data'].get('idToken'),
+                mfa_token=ret['data'].get('mfaToken')
             )
-            if self._rest._connection.consent_cache_id_token:
-                write_temporary_credential(
-                    self._rest._host, account, user, self._rest.id_token,
-                    session_parameters.get(
-                        PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL))
+            self.write_temporary_credentials(self._rest._host, user, session_parameters, ret)
             if 'sessionId' in ret['data']:
                 self._rest._connection._session_id = ret['data']['sessionId']
             if 'sessionInfo' in ret['data']:
@@ -361,80 +360,100 @@ class Auth(object):
 
         return session_parameters
 
-    def read_temporary_credential(self, host, account, user, session_parameters):
+    def _read_temporary_credential(self, host, user, cred_type):
+        cred = None
+        if IS_MACOS or IS_WINDOWS:
+            if not installed_keyring:
+                logger.debug("Dependency 'keyring' is not installed, cannot cache id token. You might experience "
+                             "multiple authentication pop ups while using ExternalBrowser Authenticator. To avoid "
+                             "this please install keyring module using the following command : pip install "
+                             "snowflake-connector-python[secure-local-storage]")
+                return
+            try:
+                cred = keyring.get_password(build_temporary_credential_name(host, user, cred_type), user.upper())
+            except keyring.errors.KeyringError as ke:
+                logger.error("Could not retrieve {} from secure storage : {}".format(cred_type, str(ke)))
+        elif IS_LINUX:
+            read_temporary_credential_file()
+            cred = TEMPORARY_CREDENTIAL.get(
+                    host.upper(), {}).get(build_temporary_credential_name(host, user, cred_type))
+        else:
+            logger.debug("OS not supported for Local Secure Storage")
+        return cred
+
+    def read_temporary_credentials(self, host, user, session_parameters):
         if session_parameters.get(PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL, False):
-            id_token = None
-            if IS_MACOS or IS_WINDOWS:
-                if not installed_keyring:
-                    # we will leave the exception for write_temporary_credential function to raise
-                    return
-                new_target = convert_target(host, user)
-                try:
-                    id_token = keyring.get_password(new_target, user.upper())
-                except keyring.errors.KeyringError as ke:
-                    logger.debug("Could not retrieve id_token from secure storage : {}".format(str(ke)))
-            elif IS_LINUX:
-                read_temporary_credential_file()
-                id_token = TEMPORARY_CREDENTIAL.get(
-                    account.upper(), {}).get(user.upper())
-            else:
-                logger.debug("connection parameter enable_sso_temporary_credential not set or OS not support")
-            self._rest.id_token = id_token
-        return
+            self._rest.id_token = self._read_temporary_credential(host, user, ID_TOKEN)
 
+        if session_parameters.get(PARAMETER_CLIENT_REQUEST_MFA_TOKEN, False):
+            self._rest.mfa_token = self._read_temporary_credential(host, user, MFA_TOKEN)
 
-def write_temporary_credential(host, account, user, id_token, store_temporary_credential=False):
-    if not id_token:
-        logger.debug("no ID token is given when try to store temporary credential")
-        return
-    if IS_MACOS or IS_WINDOWS:
-        if not installed_keyring:
-            logger.debug("Dependency 'keyring' is not installed, cannot cache id token. You might experience "
-                         "multiple authentication pop ups while using ExternalBrowser Authenticator. To avoid "
-                         "this please install keyring module using the following command : pip install "
-                         "snowflake-connector-python[secure-local-storage]")
+    def _write_temporary_credential(self, host, user, cred_type, cred):
+        if not cred:
+            logger.debug("no credential is given when try to store temporary credential")
             return
-        new_target = convert_target(host, user)
-        try:
-            keyring.set_password(new_target, user.upper(), id_token)
-        except keyring.errors.KeyringError as ke:
-            logger.debug("Could not store id_token to keyring, %s", str(ke))
-    elif IS_LINUX and store_temporary_credential:
-        write_temporary_credential_file(host, account, user, id_token)
+        if IS_MACOS or IS_WINDOWS:
+            if not installed_keyring:
+                logger.debug("Dependency 'keyring' is not installed, cannot cache id token. You might experience "
+                             "multiple authentication pop ups while using ExternalBrowser Authenticator. To avoid "
+                             "this please install keyring module using the following command : pip install "
+                             "snowflake-connector-python[secure-local-storage]")
+                return
+            try:
+                keyring.set_password(build_temporary_credential_name(host, user, cred_type), user.upper(), cred)
+            except keyring.errors.KeyringError as ke:
+                logger.error("Could not store id_token to keyring, %s", str(ke))
+        elif IS_LINUX:
+            write_temporary_credential_file(host, build_temporary_credential_name(host, user, cred_type), cred)
+        else:
+            logger.debug("OS not supported for Local Secure Storage")
+
+    def write_temporary_credentials(self, host, user, session_parameters, response):
+        if self._rest._connection.consent_cache_id_token and session_parameters.get(
+                PARAMETER_CLIENT_STORE_TEMPORARY_CREDENTIAL, False):
+            self._write_temporary_credential(host, user, ID_TOKEN, response['data'].get('idToken'))
+
+        if session_parameters.get(PARAMETER_CLIENT_REQUEST_MFA_TOKEN, False):
+            self._write_temporary_credential(host, user, MFA_TOKEN, response['data'].get('mfaToken'))
+        return
+
+
+def flush_temporary_credentials():
+    """Flush temporary credentials in memory into disk. Need to hold TEMPORARY_CREDENTIAL_LOCK."""
+    global TEMPORARY_CREDENTIAL
+    global TEMPORARY_CREDENTIAL_FILE
+    for _ in range(10):
+        if lock_temporary_credential_file():
+            break
+        time.sleep(1)
     else:
-        logger.debug("connection parameter client_store_temporary_credential not set or OS not support")
+        logger.debug("The lock file still persists after the maximum wait time."
+                     "Will ignore it and write temporary credential file: %s",
+                     TEMPORARY_CREDENTIAL_FILE)
+    try:
+        with open(TEMPORARY_CREDENTIAL_FILE, 'w',
+                  encoding='utf-8', errors='ignore') as f:
+            json.dump(TEMPORARY_CREDENTIAL, f)
+    except Exception as ex:
+        logger.debug("Failed to write a credential file: "
+                     "file=[%s], err=[%s]", TEMPORARY_CREDENTIAL_FILE, ex)
+    finally:
+        unlock_temporary_credential_file()
 
 
-def write_temporary_credential_file(host, account, user, id_token):
+def write_temporary_credential_file(host, cred_name, cred):
     """Writes temporary credential file when OS is Linux."""
     if not CACHE_DIR:
         # no cache is enabled
         return
     global TEMPORARY_CREDENTIAL
     global TEMPORARY_CREDENTIAL_LOCK
-    global TEMPORARY_CREDENTIAL_FILE
     with TEMPORARY_CREDENTIAL_LOCK:
         # update the cache
-        account_data = TEMPORARY_CREDENTIAL.get(account.upper(), {})
-        account_data[user.upper()] = id_token
-        TEMPORARY_CREDENTIAL[account.upper()] = account_data
-        for _ in range(10):
-            if lock_temporary_credential_file():
-                break
-            time.sleep(1)
-        else:
-            logger.debug("The lock file still persists. Will ignore and "
-                           "write the temporary credential file: %s",
-                           TEMPORARY_CREDENTIAL_FILE)
-        try:
-            with codecs.open(TEMPORARY_CREDENTIAL_FILE, 'w',
-                             encoding='utf-8', errors='ignore') as f:
-                json.dump(TEMPORARY_CREDENTIAL, f)
-        except Exception as ex:
-            logger.debug("Failed to write a credential file: "
-                         "file=[%s], err=[%s]", TEMPORARY_CREDENTIAL_FILE, ex)
-        finally:
-            unlock_temporary_credential_file()
+        host_data = TEMPORARY_CREDENTIAL.get(host.upper(), {})
+        host_data[cred_name.upper()] = cred
+        TEMPORARY_CREDENTIAL[host.upper()] = host_data
+        flush_temporary_credentials()
 
 
 def read_temporary_credential_file():
@@ -490,16 +509,32 @@ def unlock_temporary_credential_file():
         return False
 
 
-def delete_temporary_credential(host, user, store_temporary_credential=False):
+def delete_temporary_credential(host, user, cred_type):
     if (IS_MACOS or IS_WINDOWS) and installed_keyring:
-        new_target = convert_target(host, user)
         try:
-            keyring.delete_password(new_target, user.upper())
+            keyring.delete_password(build_temporary_credential_name(host, user, cred_type), user.upper())
         except Exception as ex:
-            logger.debug("Failed to delete credential in the keyring: err=[%s]",
-                         ex)
-    elif IS_LINUX and store_temporary_credential:
-        delete_temporary_credential_file()
+            logger.error("Failed to delete credential in the keyring: err=[%s]", ex)
+    elif IS_LINUX:
+        temporary_credential_file_delete_password(host, user, cred_type)
+
+
+def temporary_credential_file_delete_password(host, user, cred_type):
+    """Remove credential from temporary credential file when OS is Linux."""
+    if not CACHE_DIR:
+        # no cache is enabled
+        return
+    global TEMPORARY_CREDENTIAL
+    global TEMPORARY_CREDENTIAL_LOCK
+    with TEMPORARY_CREDENTIAL_LOCK:
+        # update the cache
+        host_data = TEMPORARY_CREDENTIAL.get(host.upper(), {})
+        host_data.pop(build_temporary_credential_name(host, user, cred_type), None)
+        if not host_data:
+            TEMPORARY_CREDENTIAL.pop(host.upper(), None)
+        else:
+            TEMPORARY_CREDENTIAL[host.upper()] = host_data
+        flush_temporary_credentials()
 
 
 def delete_temporary_credential_file():
@@ -516,6 +551,9 @@ def delete_temporary_credential_file():
         logger.debug("Failed to delete credential lock file: err=[%s]", ex)
 
 
-def convert_target(host, user):
-    return "{host}:{user}:{driver}".format(
-            host=host.upper(), user=user.upper(), driver=KEYRING_DRIVER_NAME)
+def build_temporary_credential_name(host, user, cred_type):
+    return "{host}:{user}:{driver}:{cred}".format(
+            host=host.upper(),
+            user=user.upper(),
+            driver=KEYRING_DRIVER_NAME,
+            cred=cred_type)
